"""
.. include:: ../README.md
"""

__all__ = [
    "alarm",
    "calendar",
    "calendar_stream",
    "event",
    "freebusy",
    "journal",
    "store",
    "timeline",
    "timespan",
    "timezone",
    "todo",
    "types",
    "tzif",
    "util",
    "diagnostics",
]
